import { ByteLength } from './common'

export class SmartBuffer {
  private _data: number[]
  private _cursor: number

  /**
  * Wrap a buffer to track position and provide useful read / write functionality.
  * @param data Buffer to wrap (optional).
  */
  constructor (data?: Uint8Array) {
    this._data = data === undefined ? [] : [...data]
    this._cursor = 0
  }

  /**
  * Return a regular buffer.
  */
  get data (): Uint8Array {
    return Uint8Array.from(this._data)
  }

  /**
  * Update the smart buffer to wrap new data.
  */
  set data (data: Uint8Array) {
    this._data = [...data]
    this.cursor = 0
  }

  /**
  * Get the length of the smart buffer data.
  */
  get length (): number {
    return this._data.length
  }

  /**
  * Set the length of the smart buffer data.
  */
  set length (length: number) {
    this._data.length = length
  }

  /**
  * Get the current cursor position of the smart buffer.
  */
  get cursor (): number {
    return this._cursor
  }

  /**
  * Set the cursor position of the smart buffer.
  */
  set cursor (position: number) {
    if (position < 0) {
      throw RangeError(`Cannot seek to ${this.cursor} of ${this.length} bytes.`)
    }
    this._cursor = position
  }

  /**
  * Pads bytes to the end of the smart buffer.
  * @param length The number of bytes to pad.
  */
  pad (length: number): void {
    this._data.push(...new Uint8Array(length))
  }

  /**
  * Return the data from the specified range.
  * @param start The start position.
  * @param end The end position.
  * @returns A new buffer containing data from the specified range.
  */
  slice (start: number, end: number): Uint8Array {
    return Uint8Array.from(this._data.slice(start, end))
  }

  /**
   * Splice bytes into the smart buffer's data.
   * @param start The position at which to insert the new data.
   * @param deleteCount The number of items to remove before inserting new data.
   * @param items The items to insert at the specified position.
   */
  splice (start: number, deleteCount: number, ...items: number[]): void {
    if (this.length < start) {
      this.pad(start)
    }
    this._data.splice(this.cursor, ByteLength.UInt16, ...items)
  }

  /**
  * Read a UInt16 number from the smart buffer at the current cursor position, and increment the cursor.
  * @returns A number represented by the bytes at the current cursor position.
  */
  readUInt16 (): number {
    this.cursor += ByteLength.UInt16
    const bytes = this.slice(this.cursor - ByteLength.UInt16, this.cursor)
    return (bytes[0] << 8) | bytes[1]
  }

  /**
  * Read a UInt32 number from the smart buffer at the current cursor position, and increment the cursor.
  * @returns A number represented by the bytes at the current cursor position.
  */
  readUInt32 (): number {
    this.cursor += ByteLength.UInt32
    const bytes = this.slice(this.cursor - ByteLength.UInt32, this.cursor)
    return (((bytes[0] << 24) | (bytes[1] << 16)) | (bytes[2] << 8) | bytes[3])
  }

  /**
  * Read the specified number of bytes from the smart buffer at the current cursor position, and increment the cursor.
  * @returns A buffer containing the bytes read from the current cursor position.
  */
  readBytes (length: number): Uint8Array {
    this.cursor += length
    return this.slice(this.cursor - length, this.cursor)
  }

  /**
  * Write a UInt16 number to the smart buffer at the current cursor position, and increment the cursor.
  * @param number The number to write in UInt16 format at the current cursor position.
  */
  writeUInt16 (number: number): void {
    const bytes = [
      (number & 0xff00) >> 8,
      (number & 0x00ff)
    ]
    this.splice(this.cursor, ByteLength.UInt16, ...bytes)
    this.cursor += ByteLength.UInt16
  }

  /**
  * Write a UInt32 number to the smart buffer at the current cursor position, and increment the cursor.
  * @param number The number to write in UInt32 format at the current cursor position.
  */
  writeUInt32 (number: number): void {
    const bytes = [
      (number & 0xff000000) >> 24,
      (number & 0x00ff0000) >> 16,
      (number & 0x0000ff00) >> 8,
      (number & 0x000000ff)
    ]
    this.splice(this.cursor, ByteLength.UInt32, ...bytes)
    this.cursor += ByteLength.UInt32
  }

  /**
  * Write the specified bytes to the smart buffer at the current cursor position, and increment the cursor.
  * @param buffer The bytes to write at the current cursor position.
  */
  writeBytes (buffer: Uint8Array): void {
    this.splice(this.cursor, buffer.length, ...buffer)
    this.cursor += buffer.length
  }
}
